/*	
	File 	: xclipboardpm.cpp
	Author	: Juergen Bohl
*/

#define INCL_WIN
#define INCL_GPI
#define INCL_DOSQUEUES
#define INCL_DOSPROCESS
#define INCL_WINWINDOWMGR
#define INCL_WINCLIPBOARD
#define INCL_DOSSEMAPHORES

#include <os2.h>
#include "XClipboard.h"
#include <string.h>

#include <stdlib.h>
#include <stdio.h>
#include <stddef.h>

#define INPUTQUEUE_TERMINATE (4242)
#define INPUTQUEUE_DATA      (42)

MRESULT EXPENTRY MyWindowProc( HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2 );
void APIENTRY InputThread(ULONG vp);
VOID Abort(HWND hwndFrame, HWND hwndClient);
void FatalError(char* text);

struct Global_ClipData
{
HMTX        hmtx;
int         size;
char        data[1];
};

HAB  hab;
PSZ  pszErrMsg;
void* global_mem=NULL;
HMTX  mutexsem=NULLHANDLE;
char* OutputQueuename;
char* SharedMemName;
int received_ntimes=0;
HWND hwnd_mainwindow=NULLHANDLE;



void setnames(int argc,char** argv)
{
int i;
char* outputqueuename,*sharedmemname;
outputqueuename = "Xclipboard";
sharedmemname   = "Xclipboard";
for (i=1;i<argc-1;i++)
    {
    if (strcmp(argv[i],"-OutputQueuename")==0)
        {
        outputqueuename = argv[i+1];
        continue;
        };
    if (strcmp(argv[i],"-SharedMemName")==0)
        {
        sharedmemname = argv[i+1];
        continue;
        };
    };
/* Set globals */
OutputQueuename = (char*)malloc(9+strlen(outputqueuename));
SharedMemName   = (char*)malloc(11+strlen(sharedmemname));
strcpy(OutputQueuename,"\\QUEUES\\");
strcpy(SharedMemName,"\\SHAREMEM\\");
strcat(OutputQueuename,outputqueuename);
strcat(SharedMemName,sharedmemname);
return;
};


int main(int argc,char** argv)
{
HMQ  hmq;
HWND hwndClient = NULLHANDLE;
HWND hwndFrame = NULLHANDLE;
QMSG qmsg;
ULONG flCreate;

setnames(argc,argv);

if ((hab = WinInitialize(0)) == 0L)
   Abort(hwndFrame, hwndClient);

if ((hmq = WinCreateMsgQueue( hab, 0 )) == 0L)
   Abort(hwndFrame, hwndClient);

if (!WinRegisterClass(
   hab,
   (PSZ)"XClipboardserver",
   (PFNWP)MyWindowProc,
   CS_SIZEREDRAW,
   0
   ))
   Abort(hwndFrame, hwndClient);

flCreate = FCF_STANDARD &
           ~FCF_SHELLPOSITION;


if ((hwndFrame = WinCreateStdWindow(
             HWND_DESKTOP,
             0,
             &flCreate,
             (PSZ)"XClipboardserver",
             (PSZ)"",
             0,
             (HMODULE)0L,
             ID_WINDOW,
             &hwndClient
             )) == 0L)
       Abort(hwndFrame, hwndClient);
hwnd_mainwindow = hwndFrame;

WinSetWindowText(hwndFrame,(PSZ)"XClipboardServer");

if (!WinSetWindowPos( hwndFrame,
                 HWND_TOP,
                 100, 100, 200, 200,
                 SWP_SIZE | SWP_MOVE | SWP_ACTIVATE | SWP_SHOW
               ))
        Abort(hwndFrame, hwndClient);

HQUEUE hqueue;
TID    tid;
DosCreateQueue(&hqueue,QUE_FIFO,(PSZ)OutputQueuename);
DosCreateThread(&tid,InputThread,(ULONG)hqueue,0,8192);

while( WinGetMsg( hab, &qmsg, 0L, 0, 0 ) )
  WinDispatchMsg( hab, &qmsg );
WinDestroyWindow(hwndFrame);

DosWriteQueue(hqueue, INPUTQUEUE_TERMINATE, 0, 0, 0);
DosWaitThread(&tid,DCWW_WAIT);
DosCloseQueue(hqueue);

WinDestroyMsgQueue( hmq );
WinTerminate( hab );

free( OutputQueuename );
free( SharedMemName );


return 0;
}

MRESULT EXPENTRY DlgWinProc(HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2)
{
switch (msg)
    {
    case WM_INITDLG :
        WinSetWindowText(WinWindowFromID(hwnd,ID_SHAREDMEMNAME),(PSZ)SharedMemName);
        WinSetWindowText(WinWindowFromID(hwnd,ID_QUEUENAME),(PSZ)OutputQueuename);
        break;
    case WM_COMMAND :
        switch (SHORT1FROMMP (mp1))
            {
            case DID_OK :
                WinDismissDlg(hwnd,DID_OK);
                break;
            };
        break;
    };
return WinDefDlgProc( hwnd, msg, mp1, mp2 );
};

MRESULT EXPENTRY MyWindowProc(HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2)
{
switch( msg )
{
  case WM_CREATE:
    BOOL b;
    APIRET rc;
    b = WinOpenClipbrd(hab);
    if (b == FALSE) break;
    b = WinSetClipbrdViewer(hab,hwnd);
    WinCloseClipbrd(hab);
    rc=DosCreateMutexSem(NULL,&mutexsem,DC_SEM_SHARED,FALSE);
    if (rc!=0) {FatalError("Couldn't open MutexSem !");break;};
    rc=DosAllocSharedMem(&global_mem,(PSZ)SharedMemName,4095,PAG_COMMIT|PAG_WRITE|PAG_READ);
    if (rc!=0) {FatalError("Couldn't allocate shared memory !");break;};
    ((Global_ClipData*)global_mem)->hmtx = mutexsem;
    ((Global_ClipData*)global_mem)->size = 4095;
    // fallthrough wanted !
  case WM_DRAWCLIPBOARD :
    {
    if (global_mem != NULL)
        {
        APIRET rc;
        HMTX hmtx = ((Global_ClipData*)global_mem)->hmtx;
        rc=DosRequestMutexSem(hmtx,100);   // Wait 100ms at most...
        if (rc==0)
            {
            if (WinOpenClipbrd(hab)==TRUE)
                {
                char* data;
                data = (char*)WinQueryClipbrdData(hab,CF_TEXT);
                if (data != NULL)
                    {
                    int stringlen,len;
                    stringlen = strlen(data)+1;
                    len = stringlen + offsetof(Global_ClipData,data);
                    if (((Global_ClipData*)global_mem)->size < len)
                        {
                        rc=DosFreeMem(global_mem);
                        if (rc!=0) {global_mem=NULL;goto goon;};   // something went wrong...
                        rc=DosAllocSharedMem(&global_mem,(PSZ)SharedMemName,len,PAG_COMMIT|PAG_WRITE|PAG_READ);
                        if (rc!=0)
                            {
                            global_mem=NULL;
                            goto goon;
                            };
                        ((Global_ClipData*)global_mem)->hmtx = mutexsem;
                        ((Global_ClipData*)global_mem)->size = len;
                        };
                    memcpy( ((Global_ClipData*)global_mem)->data, data, stringlen);
                    };
goon:           WinCloseClipbrd(hab);
                };
            DosReleaseMutexSem(hmtx);
            };
        };
    return 0;
    };
  case WM_COMMAND:
    {
    USHORT command;
    command = SHORT1FROMMP(mp1);
    switch (command)
        {
        case CM_EXIT :
            WinPostMsg(hwnd, WM_CLOSE, (MPARAM)NULL, (MPARAM)NULL);
            break;
        case CM_ABOUT :
            WinDlgBox(HWND_DESKTOP,HWND_DESKTOP,DlgWinProc,0,ID_DLG_ABOUT,NULL);
        default:
            return WinDefWindowProc( hwnd, msg, mp1, mp2 );
        }
    break;
    }
  case WM_ERASEBACKGROUND:
    return (MRESULT)( TRUE );
  case WM_PAINT:
    {
    char text[100];
    HPS    hps;
    RECTL  rc;
    POINTL pt;

    hps = WinBeginPaint( hwnd, 0L, &rc );
    WinQueryWindowRect(hwnd,&rc);
    pt.x = 5; pt.y = (rc.yTop-rc.yBottom) >> 1;
    fprintf(stdout,"%i \n",rc.yTop);
    GpiSetColor( hps, CLR_NEUTRAL );
    GpiSetBackColor( hps, CLR_BACKGROUND );
    GpiSetBackMix( hps, BM_OVERPAINT );

    if (received_ntimes == 0)
        strcpy(text,"Received no data yet");
    else
        if (received_ntimes <= 2)
            sprintf(text,"Received data %s",(received_ntimes==1)?"once":"twice");
        else
            sprintf(text,"Received data %i times",received_ntimes);
    GpiCharStringAt( hps, &pt, (LONG)strlen(text),(PSZ)text);
    WinEndPaint( hps );
    break;
    }
  case WM_CLOSE:
    if (mutexsem!=NULLHANDLE) DosCloseMutexSem(mutexsem);
    if (global_mem!=NULL) DosFreeMem(global_mem);
    WinPostMsg( hwnd, WM_QUIT, (MPARAM)0,(MPARAM)0 );
    break;
  default:
    return WinDefWindowProc( hwnd, msg, mp1, mp2 );
}
return (MRESULT)FALSE;
}

VOID Abort(HWND hwndFrame, HWND hwndClient)
{
PERRINFO  pErrInfoBlk;
PSZ       pszOffSet;
void      stdprint(void);

DosBeep(100,10);
if ((pErrInfoBlk = WinGetErrorInfo(hab)) != (PERRINFO)NULL)
    {
    pszOffSet = ((PSZ)pErrInfoBlk) + pErrInfoBlk->offaoffszMsg;
    pszErrMsg = ((PSZ)pErrInfoBlk) + *((PSHORT)pszOffSet);
    if((INT)hwndFrame && (INT)hwndClient)
      WinMessageBox(HWND_DESKTOP,         /* Parent window is desk top */
                    hwndFrame,            /* Owner window is our frame */
                    (PSZ)pszErrMsg,       /* PMWIN Error message       */
                    (PSZ)"Error Msg",     /* Title bar message         */
                    MSGBOXID,             /* Message identifier        */
                    MB_MOVEABLE | MB_CUACRITICAL | MB_CANCEL ); /* Flags */
    WinFreeErrorInfo(pErrInfoBlk);
    };
WinPostMsg(hwndClient, WM_QUIT, (MPARAM)NULL, (MPARAM)NULL);
};

void FatalError(char* text)
{
WinMessageBox(HWND_DESKTOP,
              hwnd_mainwindow,
              (PSZ)text,
              (PSZ)"Error Msg",
              MSGBOXID,
              MB_MOVEABLE | MB_CUACRITICAL | MB_CANCEL );
WinPostMsg(hwnd_mainwindow, WM_QUIT, (MPARAM)NULL, (MPARAM)NULL);
};

void APIENTRY InputThread(ULONG h)
{
HAB hab = WinInitialize(0);
HQUEUE hqueue = (HQUEUE)h;
ULONG len;
char* data;
REQUESTDATA reqdat;
BYTE prio;
APIRET rc;
for (;;)
    {
    rc=DosReadQueue(hqueue,&reqdat,&len,(PPVOID)&data,0,0,&prio,0);
    if (rc==0)
        {
        if (reqdat.ulData == INPUTQUEUE_DATA)
            {
            if (WinOpenClipbrd(hab)==FALSE)
                break;
            else
                {
                WinSetClipbrdData(hab,(ULONG)data,CF_TEXT,CFI_POINTER);
                WinCloseClipbrd(hab);
                received_ntimes++;
                WinInvalidateRect(hwnd_mainwindow,NULL,TRUE);
                };
            }
            else if (reqdat.ulData == INPUTQUEUE_TERMINATE) break;
        }
    else
        break;
    };
WinTerminate(hab);
DosExit(EXIT_THREAD,0);
};
